#include "filed.h"
#include "externs.h"
#include "peer.h"
#include "socket.h"
#include "main.h"
#include "log.h"
#include "io.h"
#include "parse.h"
#include "queue.h"

#include <time.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>

void handle_peers(void) {
	/* walk through connection pool */
	/* for any peers we haven't talked to in heartbeat seconds,
	   send heartbeat */

	/* for any peers that we don't have a connection to, initiate a connection */

	struct connection *c;
	time_t now = time(NULL);
	int peer_hb;
	int peer_timeout;

	peer_hb = get_one_data_int(filed->cf,"peer-hb");
	peer_timeout = get_one_data_int(filed->cf,"peer-timeout");
	
	c = filed->con_head;

	while(c) {
		
		if(c->type == CTYPE_PEER) {
			if(((c->last_msg + peer_hb) < now) && ((c->last_msg + peer_timeout) > now)) 
			peer_send_heartbeat(c);
			if((c->last_recv + peer_timeout) < now)
				peer_disconnect(c);
		}
	c = c->next;
	}

}

void peer_send_heartbeat(struct connection *c) {
	if(!strcmp(c->name,"")) return;
	c_send(c,"HB %d %d %d",filed->load,filed->s_count,filed->peer_flags);
}

void peer_disconnect(struct connection *c) {
	c->err = ins_queue(c->err,NULL,QFLAG_DISCONNECT);
}	

void determine_best_peer(void) {
	/*
	 * we determine the best peer by looking at peer data
	 * collected during heartbeat
	 *
	 */
	int id;
	int meta;
	int tmeta;
	int prev_peer;

	struct connection *c = filed->con_head;
	
	id = 0;
	meta = (filed->load * 100) + filed->s_count;
	if(meta > 100) 
		meta -= 100;
		else
		meta = 0;

	if(filed->peer_flags & PFLAG_DOWN) meta = 32767;
	
	prev_peer = filed->best_peer;
	filed->best_peer = 0;

	while(c) {
		if(c->type == CTYPE_PEER) {
			if(!(c->flags & CFLAG_NODATA) && !(c->pflags & PFLAG_DOWN)) {
				tmeta = c->s_count + (c->load * 100);
				if(tmeta < meta) {
					meta = tmeta;
					id = c->id;
				}
			}
		}
	c = c->next;
	}

	filed->best_peer = id;
	if(id != prev_peer) dlog(DEBUG_PEER,"Changed best peer from %d to %d",prev_peer,id);

return;
}

void reconnect_peers(void) {
	struct parse *p;
	struct parse_hdr *p_hdr;
	int port;
	char *name;
	char *cport;
	char *ip;
	struct connection *c;
	time_t now;
	int rc;

	rc = get_one_data_int(filed->cf,"peer-reconnect");

	now = time(NULL);
	if(now % rc) return;

	p_hdr = filed->cf;

	while((p_hdr = get_entry(p_hdr,"peer"))) {
		p = p_hdr->data;
		
		/* get name, ip, port */
		/* attempt to connect */
		name = get_token_s(p,1);
		ip = get_token_s(p,2);
		cport = get_token_s(p,3);

		if(!name || (strlen(name) > 20)) {
			p_hdr = p_hdr->next;
			continue;
		}

		if(!strcasecmp(name,filed->hostname)) {
			p_hdr = p_hdr->next;
			continue;
			/* 
			 * you can't talk to yourself. People 
			 * would think you were crazy .. 
			 *
			 */
		}

		if(!ip) {
			p_hdr = p_hdr->next;
			continue;
		}

		if(!cport) {
			port = get_one_data_int(filed->cf,"port");
		} else {
			port = atoi(cport);
		}
		
		if(peer_get_by_name(name)) {
			p_hdr = p_hdr->next;
			continue;
		}

		dlog(DEBUG_PEER,"reconnect_peers(): connecting to [%s] %s/%d",name,ip,port);

		if(!(c = connect_to(ip,port))) {
			p_hdr = p_hdr->next;
			dlog(DEBUG_PEER,"reconnect_peers(): connect to [%s] %s/%d failed",name,ip,port);
			continue;
		}

		c->type = CTYPE_PEER;
		c->load = 0;
		c->s_count = 0;
		c->flags = CFLAG_NODATA;
	
		insert_connection(c);
		dlog(DEBUG_PEER,"reconnect_peers(): [%d] connect to [%s] %s/%d done first stage",c->id,name,ip,port);
	
		/* if connection succeeds, insert connection into queue */

		p_hdr = p_hdr->next;
	
		/*
		 * when parse_peer gets a OK back 
		 * and sees it has no name for this host, 
		 * it will send a whoareyou
		 * 
		 * if it can't find a peer with that name
		 * or if there's already a link up
		 * it will send an error and disconnect
		 *
		 */

		}
}

void peer_become(struct connection *c, struct parse *p) {
	char *name;

	name = get_token_s(p,1);
	
	if(!name) {
		c_send(c,"ERR ULINK \"PEER requires an argument\"");
		c_hangup(c);
	}
	
	strcpy(c->name,"");
	if(peer_get_by_name(name)) {
		c_send(c,"ERR ULINK \"PEER [%s] is already connected\"");
		c_hangup(c);
	}
	if(!strcasecmp(name,filed->hostname)) {
		c_send(c,"ERR ULINK \"This town's only big enough for one of us\"");
		c_hangup(c);
	}

	strcpy(c->name,name);
	c->type = CTYPE_PEER;
	c->flags += CFLAG_NODATA;
	return;
}

void peer_process_ok(struct connection *c, struct parse *p) {
	if(!strcmp(c->name,"")) {
		c_send(c,"PEER \"%s\"",filed->hostname);
	} else {
		/* do nothing */
	}
	return;
}

void peer_process_err(struct connection *c, struct parse *p) {
	char *arg1;
	char *ch;
	char arg2[1024];
	int i=2;

	arg1 = get_token_s(p,1);
	strcpy(arg2,"");

	while((ch = get_token_s(p,i++)))
		strcat(arg2,ch);
	
	dlog(DEBUG_PEER,"peer %s reported error: %s",c->name,arg2);
	
	if(!strcasecmp(arg1,"ULINK")) c_hangup(c);
}
	
void peer_process_peer(struct connection *c, struct parse *p) {
	if(c->name && strcmp(c->name,"")) {
		dlog(DEBUG_CODE,"peer_process_peer(): peer_become() already has %s loaded",c->name);
		return;
	}
	peer_become(c,p);
	c_send(c,"PEER \"%s\"",filed->hostname);
	return;
}

void peer_process_hb(struct connection *c, struct parse *p) {
	char *load;
	char *ses;
	char *flag;

	load = get_token_s(p,1);
	ses = get_token_s(p,2);
	flag = get_token_s(p,3);
	
	if(!load || !ses || !flag) {
		c_send(c,"ERR ULINK \"Incorrect number of arguments for HB\"");
	} else {

		c->load = atoi(load);
		c->s_count = atoi(ses);
		c->pflags = atoi(flag);
		if(c->flags & CFLAG_NODATA) c->flags -= CFLAG_NODATA;
	}
	return;
}
	

struct connection *peer_get_by_name(char *name) {
	struct connection *c = filed->con_head;

	while(c) {
		if(c->type == CTYPE_PEER) {
			if(!strcasecmp(c->name,name))
				return(c);
		}
	c = c->next;
	}
	return(NULL);
}

struct connection *peer_get_by_id(int id) {
	struct connection *c = filed->con_head;
	
	while(c) {
		if(c->type == CTYPE_PEER) {
			if(c->id == id)
				return(c);
		}
	c = c->next;
	}

	return(NULL);
}

void peer_connection_dump(struct connection *c, struct parse *p) {
	int i;
	struct connection *d = filed->con_head;
	char *ch;
	int found = 0;
	

	if(get_token_s(p,2)) {

		while(d) {
			if(d->type == CTYPE_PEER) {
				i = 2;
				while((ch = get_token_s(p,i++))) {
					if(!strcasecmp(ch,d->name)) {
						c_send(c,"peer [%s]: %d %d %s",d->name,d->load,d->s_count,peer_get_pflags_string(d->pflags));
						c_send(c,"%s %s",ch,d->name);
						found++;
					}
				}
			}
		d = d->next;
		}

	} else {

		while(d) {
			if(d->type == CTYPE_PEER) {
				c_send(c,"peer [%s]: %d %d %s",d->name,d->load,d->s_count,peer_get_pflags_string(d->pflags));
				found++;
			}
		d = d->next;
		}

	}
	c_send(c,"%d peers found");
	return;
}

char *peer_get_pflags_string(int flags) {
	static char ret[1024];

	strcpy(ret,"");
	if(flags & PFLAG_DOWN) strcat(ret,"DOWN ");
	if(flags & PFLAG_ADMINDOWN) strcat(ret,"ADMINDOWN ");

	return(ret);
}

int peer_proxy_to(int peer_id, char *fmt, ...) {
	char outbuf[1024];
	va_list ap;

	va_start(ap,fmt);
	vsprintf(outbuf,fmt,ap);
	va_end(ap);

	if(!peer_get_by_id(peer_id)) return(0);
	dlog(DEBUG_PEER,"Proxying %s to %d",outbuf,peer_id);
	c_send_to(peer_id,outbuf);
	
	return(1);
}

void peer_process_tx(struct connection *c, struct parse *p) {
	char *con;
	char *msg;
	int c_id;
	
	con = get_token_s(p,1);
	msg = get_token_s(p,2);

	if(!con || !msg) {
		c_send(c,"ERR ULINK \"TX requires a connection and a message\"");
		return;
	}

	c_id = atoi(con);
	c_send_to(c_id,msg);
}
